from torch.utils.data import Dataset, DataLoader
import numpy as np
import scanpy as sc


class CSDataSet(Dataset):
    def __init__(self, ctrl, stim):
        '''
        Build dataset of control and stimulated adata
        :param ctrl: control adata
        :param stim: stimulated adata
        '''
        self.ctrl = ctrl.to_df().values
        self.stim = stim.to_df().values

    def __getitem__(self, index):
        x = self.ctrl[index, :]
        y = self.stim[index, :]
        return x, y

    def __len__(self):
        return self.ctrl.shape[0]


class AnnDataSet(Dataset):
    def __init__(self, adata):
        '''
        Build dataset of adata
        :param adata: adata of training or testing set
        '''
        self.data = adata.to_df().values

    def __getitem__(self, index):
        x = self.data[index, :]
        return x

    def __len__(self):
        return self.data.shape[0]


def data_loader(adata, key_dic, bz=128, cell_to_pred='CD4T'):
    '''
    Build the training set from the input adata
    :param adata: complete adata
    :param key_dic: dictionary of keywords for the data set
    :param bz: batch size
    :param cell_to_pred: cell type to predict
    :return: training set
    '''
    cell_type_key = key_dic['cell_type_key']
    condition_key = key_dic['condition_key']
    ctrl_key = key_dic['ctrl_key']
    stim_key = key_dic['stim_key']
    types = list(set(adata.obs[cell_type_key]))
    new_adata = adata[(adata.obs[cell_type_key] == 'none')]
    for cell_type in types:
        if cell_type != cell_to_pred:
            ctrl = adata[((adata.obs[cell_type_key] == cell_type) &
                          (adata.obs[condition_key] == ctrl_key))]
            stim = adata[((adata.obs[cell_type_key] == cell_type) &
                          (adata.obs[condition_key] == stim_key))]

            ctrl_ind = np.random.choice(range(ctrl.shape[0]), size=(int(ctrl.shape[0] / bz) + 1) * bz)
            stim_ind = np.random.choice(range(stim.shape[0]), size=(int(stim.shape[0] / bz) + 1) * bz)
            ctrl_adata = ctrl[ctrl_ind, :]
            stim_adata = stim[stim_ind, :]
            ctrl_adata.obs["label"] = [cell_type + '_ctrl'] * ctrl_adata.obs.shape[0]
            stim_adata.obs["label"] = [cell_type + '_stim'] * stim_adata.obs.shape[0]
            if new_adata.obs.shape[0] == 0:
                new_adata = ctrl_adata.concatenate(stim_adata)
            else:
                new_adata = new_adata.concatenate(ctrl_adata, stim_adata)
        else:
            ctrl = adata[((adata.obs[cell_type_key] == cell_type) &
                          (adata.obs[condition_key] == ctrl_key))]
            ctrl_ind = np.random.choice(range(ctrl.shape[0]), size=(int(ctrl.shape[0] / bz) + 1) * bz)
            ctrl_adata = ctrl[ctrl_ind, :]
            ctrl_adata.obs["label"] = [cell_type + '_ctrl'] * ctrl_adata.obs.shape[0]
            if new_adata.obs.shape[0] == 0:
                new_adata = ctrl_adata
            else:
                new_adata = new_adata.concatenate(ctrl_adata)
    train_set = AnnDataSet(new_adata)
    train_loader = DataLoader(dataset=train_set, batch_size=bz, shuffle=False,
                              num_workers=0, drop_last=False)
    return train_loader


if __name__ == '__main__':
    adata = sc.read_h5ad('../data/pbmc_perturbed.h5ad')
    input_dim = adata.var.shape[0]
    cell_to_pred = 'CD4T'
    key_dic = {'condition_key': 'condition',
               'cell_type_key': 'cell_type',
               'ctrl_key': 'control',
               'stim_key': 'stimulated'}

    train_loader = data_loader(adata, key_dic, bz=128, cell_to_pred='CD4T')


import os
import numpy as np
import scanpy as sc


def get_less_adata(adata, key_dic, frac=0.2):
    """
    Dataset downsampling
    :param adata: input adata
    :param key_dic: dictionary of keywords for the dataset, such as
               {'condition_key': 'condition',
               'cell_type_key': 'cell_type',
               'ctrl_key': 'control',
               'stim_key': 'stimulated',
               'pred_key': 'pred',
               }
    :param frac: downsampling rate
    :return: New dataset obtained by downsampling
    """
    if (frac > 1) | (frac < 0):
        print("Info: the fac should be in [0,1]")
        return None
    types = set(adata.obs[key_dic['cell_type_key']])
    conditions = set(adata.obs[key_dic['condition_key']])
    res = {}
    for i in range(int(1 / (frac + 0.001)) + 1):
        res['part{}'.format(i)] = []
    for cell_to_pred in types:
        for condition in conditions:
            adata_new = adata[((adata.obs[key_dic['cell_type_key']] == cell_to_pred) &
                               (adata.obs[key_dic['condition_key']] == condition))]
            size = int(frac * adata_new.n_obs)
            idxs = np.array(range(adata_new.n_obs))
            np.random.shuffle(idxs)
            for i in range(int(1 / (frac + 0.001))):
                idx = idxs[0:size * (i + 1)]
                res['part{}'.format(i)].append(adata_new[idx])
            res['part{}'.format(int(1 / (frac + 0.001)))].append(adata_new)
    for i in range(int(1 / (frac + 0.001)) + 1):
        res['part{}'.format(i)] = sc.AnnData.concatenate(*res['part{}'.format(i)], join='outer')
    return res


def check_dir(path='./'):
    """
    Check that project folders exist for the specified directory, and create them if not
    :param path: the specified directory
    :return: None
    """
    img_path = path + 'images'
    log_path = path + 'log'
    res_path = path + 'results'
    mod_path = path + "save_models"

    if not os.path.isdir(img_path):
        os.mkdir(img_path)
    if not os.path.isdir(log_path):
        os.mkdir(log_path)
    if not os.path.isdir(res_path):
        os.mkdir(res_path)
    if not os.path.isdir(mod_path):
        os.mkdir(mod_path)


# 用于平衡各个类型的数目, 按照最多进行补齐或者最少进行采样
def balancer(adata, type_key, max=True):
    """
    Balance the number of cells of each type, with maximum replenishment or minimum sampling
    :param adata: the input adata
    :param type_key: the key of cell type in adata.obs
    :param max: maximum if True and minimum if False
    :return: the balanced data
    """
    class_names = np.unique(adata.obs[type_key])
    class_pop = {}
    for cls in class_names:
        class_pop[cls] = adata[adata.obs[type_key] == cls].shape[0]
    if max:
        number = np.max(list(class_pop.values()))
    else:
        number = np.min(list(class_pop.values()))
    index_all = []
    for cls in class_names:
        class_index = np.array(adata.obs[type_key] == cls)
        index_cls = np.nonzero(class_index)[0]
        index_cls_r = index_cls[np.random.choice(len(index_cls), number)]
        index_all.append(index_cls_r)

    balanced_data = adata[np.concatenate(index_all)].copy()
    return balanced_data


def get_ctrl_stim(latent_adata, key_dic):
    """
    Get the control and perturbed latent data after balance
    :param latent_adata: the input latent adata
    :param key_dic: dictionary of keywords for the dataset, such as
               {'condition_key': 'condition',
               'cell_type_key': 'cell_type',
               'ctrl_key': 'control',
               'stim_key': 'stimulated',
               'pred_key': 'pred',
               }
    :return: the control and perturbed latent data after balance
    """
    condition_key = key_dic['condition_key']
    cell_type_key = key_dic['cell_type_key']
    ctrl_key = key_dic['ctrl_key']
    stim_key = key_dic['stim_key']
    ctrl_x = latent_adata[latent_adata.obs[condition_key] == ctrl_key, :]
    stim_x = latent_adata[latent_adata.obs[condition_key] == stim_key, :]
    ctrl_x = balancer(ctrl_x, cell_type_key)
    stim_x = balancer(stim_x, cell_type_key)
    eq = min(ctrl_x.X.shape[0], stim_x.X.shape[0])
    cd_ind = np.random.choice(range(ctrl_x.shape[0]), size=eq, replace=False)
    stim_ind = np.random.choice(range(stim_x.shape[0]), size=eq, replace=False)
    ctrl_z = ctrl_x[cd_ind, :]
    stim_z = stim_x[stim_ind, :]
    return ctrl_z, stim_z


# 获取对照组的数据
def get_ctrl_adata(adata, cell_to_pred, key_dic):
    """
    Get the control adata of certain cell type
    :param adata: the input adata
    :param cell_to_pred: the cell type that you want to predict
    :param key_dic: key_dic: dictionary of keywords for the dataset, such as
               {'condition_key': 'condition',
               'cell_type_key': 'cell_type',
               'ctrl_key': 'control',
               'stim_key': 'stimulated',
               'pred_key': 'pred',
               }
    :return: control adata
    """
    ctrl_adata = adata[((adata.obs[key_dic['cell_type_key']] == cell_to_pred) &
                        (adata.obs[key_dic['condition_key']] == key_dic['ctrl_key']))]
    return ctrl_adata






import torch
import torch.nn as nn
from torch import Tensor, optim
from tqdm import tqdm
from scpram.dataset import AnnDataSet
from scpram.utils import balancer
from torch.utils.data import DataLoader
import scanpy as sc
from torch.distributions import Normal
from torch.distributions import kl_divergence as kl
import numpy as np
from sklearn.metrics.pairwise import cosine_similarity
from sklearn.preprocessing import normalize
import ot


class SCPRAM(nn.Module):
    def __init__(self, input_dim=6998, latent_dim=100, hidden_dim=1000,
                 noise_rate=0.1, kl_weight=5e-4, device=None):
        '''
        scPRAM model
        :param input_dim: number of input genes
        :param latent_dim: the number of latent space neurons
        :param hidden_dim: the number of hidden space neurons
        :param noise_rate: the amount of data noise
        :param kl_weight: weight of KL divergence in loss function
        :param device: hardware equipment for training models
        '''
        super(SCPRAM, self).__init__()
        self.input_dim = input_dim
        self.hidden_dim = hidden_dim
        self.latent_dim = latent_dim
        self.noise_rate = noise_rate
        self.kl_weight = kl_weight
        self.device = torch.device('cuda' if torch.cuda.is_available() else 'cpu') if device is None else device

        # Encoder layers
        self.encoder = nn.Sequential(
            nn.Linear(input_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, latent_dim * 2)
        )

        # Decoder layers
        self.decoder = nn.Sequential(
            nn.Linear(latent_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, hidden_dim),
            nn.ReLU(),
            nn.Linear(hidden_dim, input_dim),
            nn.ReLU()
        )

    def encode(self, x):
        h = self.encoder(x)
        mu, logvar = torch.chunk(h, 2, dim=1)
        z = self.reparameterize(mu, logvar)
        return z, mu, logvar

    def decode(self, z):
        x_hat = self.decoder(z)
        return x_hat

    def reparameterize(self, mu, logvar):
        std = torch.exp(0.5 * logvar)
        eps = torch.randn_like(std)
        z = mu + eps * std
        return z

    def forward(self, x):
        noise = torch.randn_like(x)
        x_noisy = x + noise * self.noise_rate
        z, mu, logvar = self.encode(x_noisy)
        x_hat = self.decode(z)

        std = torch.exp(logvar / 2)
        loss_kl = kl(
            Normal(mu, std),
            Normal(0, 1)
        ).sum(dim=1)
        loss_rec = ((x - x_hat) ** 2).sum(dim=1)

        return x_hat, loss_rec, loss_kl

    def get_latent_adata(self, adata):
        device = torch.device('cuda' if torch.cuda.is_available() else 'cpu')
        x = Tensor(adata.to_df().values).to(device)
        latent_z = self.encode(x)[0].cpu().detach().numpy()
        latent_adata = sc.AnnData(X=latent_z, obs=adata.obs.copy())
        return latent_adata

    def get_loss(self, x):
        x_hat, loss_rec, loss_kl = self.forward(x)
        return loss_rec, loss_kl

    def train_SCPRAM(self, train_adata, epochs=100, batch_size=128, lr=5e-4):
        '''
        train the SCPRAM
        :param train_adata: adata of training set
        :param epochs: the number of training epochs
        :param batch_size: the size of each batch of training data
        :param lr: learning rate
        :return: None
        '''
        device = self.device
        pbar = tqdm(range(epochs))
        anndataset = AnnDataSet(train_adata)
        train_loader = DataLoader(anndataset, batch_size=batch_size, shuffle=True, drop_last=False)
        SCPRAM_loss, loss_rec, loss_kl = 0, 0, 0
        optim_SCPRAM = optim.Adam(self.parameters(), lr=lr, weight_decay=1e-5)
        for epoch in pbar:
            pbar.set_description("Training Epoch {}".format(epoch))
            for idx, x in enumerate(train_loader):
                x = x.to(device)
                loss_rec, loss_kl = self.get_loss(x)
                SCPRAM_loss = (0.5 * loss_rec + 0.5 * (loss_kl * self.kl_weight)).mean()
                optim_SCPRAM.zero_grad()
                SCPRAM_loss.backward()
                optim_SCPRAM.step()
            pbar.set_postfix(SCPRAM_loss=SCPRAM_loss.item(), recon_loss=loss_rec.mean().item(),
                             kl_loss=loss_kl.mean().item())

    def ot_predict(self, adata_train, cell_to_pred, key_dic):
        '''
        The optimal transport mapping is used directly to predict the expression state after perturbation
        This method has some effect, but is not accurate enough.
        :param adata_train: adata of training set
        :param cell_to_pred: cell type to predict
        :param key_dic: dictionary of keywords for the data set
        :return: the predicted response after the perturbation
        '''
        ctrl_adata = adata_train[(adata_train.obs[key_dic['condition_key']] == key_dic['ctrl_key'])]
        z_train = self.get_latent_adata(adata_train)
        z_ctrl_adata = z_train[(z_train.obs[key_dic['condition_key']] == key_dic['ctrl_key'])]
        z_stim_adata = z_train[(z_train.obs[key_dic['condition_key']] == key_dic['stim_key'])]
        z_ctrl = z_ctrl_adata.to_df().values
        z_stim = z_stim_adata.to_df().values

        M = ot.dist(z_ctrl, z_stim, metric='euclidean')
        G = ot.emd(torch.ones(z_ctrl.shape[0]) / z_ctrl.shape[0],
                   torch.ones(z_stim.shape[0]) / z_stim.shape[0],
                   torch.tensor(M), numItermax=1000000)
        z_pred = torch.mm(G, torch.tensor(z_stim)).numpy() * z_ctrl.shape[0]

        pred_x = self.decode(Tensor(z_pred).to(self.device)).cpu().detach().numpy()
        pred_adata = sc.AnnData(X=pred_x, obs=ctrl_adata.obs.copy(), var=ctrl_adata.var.copy())
        pred = pred_adata[(pred_adata.obs[key_dic['cell_type_key']]) == cell_to_pred]
        pred.obs[key_dic['condition_key']] = key_dic['pred_key']
        return pred

    def cross_cell_predict(self, train_adata, cell_to_pred, key_dic, n_top=None):
        '''
        Cross-cell type prediction, that is, using pre - and post-perturbation data for the first N known cell types,
        predicts the post-perturbation response of the N+1 cell type. This approach requires the use of cell labels
        and does not require the use of OT to match cells. The prediction worked well.
        :param train_adata: adata of training set
        :param cell_to_pred: cell type to predict
        :param key_dic: dictionary of keywords for the dataset
        :param n_top: what percentage of cells are used for attention mechanisms
        :return: the predicted response after the perturbation
        '''
        ctrl_adata = train_adata[((train_adata.obs[key_dic['cell_type_key']] == cell_to_pred) &
                                  (train_adata.obs[key_dic['condition_key']] == key_dic['ctrl_key']))]
        train_z = self.get_latent_adata(train_adata)
        ctrl_z = train_z[(train_z.obs[key_dic['condition_key']] == key_dic['ctrl_key'])]
        stim_z = train_z[(train_z.obs[key_dic['condition_key']] == key_dic['stim_key'])]
        print(ctrl_z.shape, stim_z.shape)
        ctrl_z = balancer(ctrl_z, key_dic['cell_type_key'])
        stim_z = balancer(stim_z, key_dic['cell_type_key'])
        eq = min(ctrl_z.X.shape[0], stim_z.X.shape[0])
        cd_ind = np.random.choice(range(ctrl_z.shape[0]), size=eq, replace=False)
        stim_ind = np.random.choice(range(stim_z.shape[0]), size=eq, replace=False)
        ctrl_z = ctrl_z[cd_ind, :]
        stim_z = stim_z[stim_ind, :]
        types = set(train_adata.obs[key_dic['cell_type_key']])
        ctrl_list, stim_list = [], []
        for cell_type in types:
            if (cell_type == cell_to_pred) or (cell_type == 'isolated'):
                continue
            ctrl_m = ctrl_z[(ctrl_z.obs[key_dic['cell_type_key']] == cell_type)].to_df().values.mean(axis=0)
            stim_m = stim_z[(stim_z.obs[key_dic['cell_type_key']] == cell_type)].to_df().values.mean(axis=0)
            if len(ctrl_list) > 0:
                ctrl_list = np.vstack((ctrl_list, ctrl_m))
                stim_list = np.vstack((stim_list, stim_m))
            else:
                ctrl_list = ctrl_m
                stim_list = stim_m

        delta_list = np.array(stim_list - ctrl_list).reshape(-1, self.latent_dim)
        test_z = self.get_latent_adata(ctrl_adata).to_df().values
        # calulate the cosine similarity
        cos_sim = cosine_similarity(np.array(test_z).reshape(-1, self.latent_dim),
                                    np.array(ctrl_list).reshape(-1, self.latent_dim))
        if n_top is None:
            cos_sim = normalize(cos_sim, axis=1, norm='l1').reshape(test_z.shape[0], -1)
            delta_pred = np.matmul(cos_sim, delta_list)
        else:
            top_indices = np.argsort(cos_sim)[0][-n_top:]
            normalized_weights = cos_sim[0][top_indices] / np.sum(cos_sim[0][top_indices])
            delta_pred = np.sum(normalized_weights[:, np.newaxis] *
                                np.array(delta_list).reshape(-1, self.latent_dim)[top_indices], axis=0)
        pred_z = test_z + delta_pred
        pred_x = self.decode(Tensor(pred_z).to(self.device)).cpu().detach().numpy()
        pred_adata = sc.AnnData(X=pred_x, obs=ctrl_adata.obs.copy(), var=ctrl_adata.var.copy())
        pred_adata.obs[key_dic['condition_key']] = key_dic['pred_key']
        return pred_adata

    def predict(self, train_adata, cell_to_pred, key_dic, ratio=0.05):
        '''
        Method introduced in the paper. OT is used to match the cells before and after perturbation,
        and then the perturbation vector is solved using the attention mechanism.
        :param train_adata: adata of training set including the control data of type to predict
        :param cell_to_pred: cell type to predict
        :param key_dic: dictionary of keywords for the data set
        :param ratio: what percentage of cells are used for attention mechanisms
        :return: the predicted response after the perturbation
        '''
        ctrl_to_pred = train_adata[((train_adata.obs[key_dic['cell_type_key']] == cell_to_pred) &
                                    (train_adata.obs[key_dic['condition_key']] == key_dic['ctrl_key']))]
        ctrl_adata = train_adata[(train_adata.obs[key_dic['cell_type_key']] != cell_to_pred) &
                                 (train_adata.obs[key_dic['condition_key']] == key_dic['ctrl_key'])]
        stim_adata = train_adata[(train_adata.obs[key_dic['condition_key']] == key_dic['stim_key'])]

        ctrl = self.get_latent_adata(ctrl_adata).to_df().values
        stim = self.get_latent_adata(stim_adata).to_df().values
        M = ot.dist(stim, ctrl, metric='euclidean')
        G = ot.emd(torch.ones(stim.shape[0]) / stim.shape[0],
                   torch.ones(ctrl.shape[0]) / ctrl.shape[0],
                   torch.tensor(M), numItermax=100000)
        match_idx = torch.max(G, 0)[1].numpy()
        stim_new = stim[match_idx]
        delta_list = stim_new - ctrl
        test_z = self.get_latent_adata(ctrl_to_pred).to_df().values
        cos_sim = cosine_similarity(np.array(test_z).reshape(-1, self.latent_dim),
                                    np.array(ctrl).reshape(-1, self.latent_dim))
        n_top = int(np.ceil(ctrl.shape[0] * ratio))
        top_indices = np.argsort(cos_sim)[0][-n_top:]
        normalized_weights = cos_sim[0][top_indices] / np.sum(cos_sim[0][top_indices])
        delta_pred = np.sum(normalized_weights[:, np.newaxis] *
                            np.array(delta_list).reshape(-1, self.latent_dim)[top_indices], axis=0)
        pred_z = test_z + delta_pred
        pred_x = self.decode(Tensor(pred_z).to(self.device)).cpu().detach().numpy()
        pred_adata = sc.AnnData(X=pred_x, obs=ctrl_to_pred.obs.copy(), var=ctrl_to_pred.var.copy())
        pred_adata.obs[key_dic['condition_key']] = key_dic['pred_key']
        return pred_adata



import seaborn as sns
import pandas as pd
import scanpy as sc
import numpy as np
import matplotlib.pyplot as plt
from adjustText import adjust_text
from scipy.stats import wasserstein_distance
from matplotlib.backends.backend_pdf import PdfPages
import sklearn
import ot


def evaluate_adata(eval_adata, cell_type, key_dic):
    """
    Evaluate predictive performance; Note that this function is used when one of the subexperiments
    on a dataset have been completed
    :param cell_type: the cell type that you are evaluating
    :param eval_adata: adata of gene expression containing ctrl,stim, pred
    :param key_dic: dictionary of keywords for the dataset, such as
               {'condition_key': 'condition',
               'cell_type_key': 'cell_type',
               'ctrl_key': 'control',
               'stim_key': 'stimulated',
               'pred_key': 'pred',
               }
    :return: None
    """
    # PCA cluster results
    sc.tl.pca(eval_adata)
    sc.pl.pca(eval_adata, color=key_dic['condition_key'], frameon=False,
              title="PCA of " + cell_type + " by Condition")
    # conmmon degs of top 100
    sc.tl.rank_genes_groups(eval_adata, groupby=key_dic['condition_key'],
                            reference=key_dic['ctrl_key'], method="wilcoxon")
    degs_pred = eval_adata.uns["rank_genes_groups"]["names"][key_dic['pred_key']]
    degs_ctrl = eval_adata.uns["rank_genes_groups"]["names"][key_dic['stim_key']]
    common_degs = list(set(degs_ctrl[0:100]) & set(degs_pred[0:100]))
    common_nums = len(common_degs)
    print("common DEGs: ", common_nums)
    # regression plot of mean
    draw_reg_plot(eval_adata=eval_adata,
                  cell_type=cell_type,
                  reg_type='mean',
                  axis_keys={"x": "pred", "y": key_dic['stim_key']},
                  condition_key=key_dic['condition_key'],
                  gene_draw=degs_ctrl[:10],
                  top_gene_list=degs_ctrl[:100],
                  save_path=None,
                  title=None,
                  show=True,
                  return_fig=False,
                  fontsize=12
                  )
    # regression plot of var
    draw_reg_plot(eval_adata=eval_adata,
                  cell_type=cell_type,
                  reg_type='var',
                  axis_keys={"x": "pred", "y": key_dic['stim_key']},
                  condition_key=key_dic['condition_key'],
                  gene_draw=degs_ctrl[:10],
                  top_gene_list=degs_ctrl[:100],
                  save_path=None,
                  title=None,
                  show=True,
                  return_fig=False,
                  fontsize=12
                  )
    # violin contrast
    gene = degs_ctrl[0]
    sc.pl.violin(eval_adata, keys=gene, groupby=key_dic['condition_key'])
    gene = degs_ctrl[1]
    sc.pl.violin(eval_adata, keys=gene, groupby=key_dic['condition_key'])
    gene = degs_ctrl[2]
    sc.pl.violin(eval_adata, keys=gene, groupby=key_dic['condition_key'])
    # degs contrast
    sc.tl.rank_genes_groups(eval_adata, groupby=key_dic['condition_key'],
                            reference=key_dic['ctrl_key'], method="wilcoxon")
    sc.pl.rank_genes_groups(eval_adata, n_genes=25, sharey=False, show=True)

    marker_genes = degs_ctrl[0:20]
    sc.pl.dotplot(eval_adata, marker_genes, groupby=key_dic['condition_key'], show=True)


def evaluate(data_name, eval_path, key_dic, save_path=None, return_fig=False):
    '''
    Evaluate predictive performance; Note that this function is used when all the subexperiments
    on a dataset have been completed
    :param data_name: name of dataset, such as PBMC
    :param eval_path: the path where you save the results of prediction
    :param key_dic: dictionary of keywords for the dataset
    :param save_path: the path to save the evaluation results
    :param return_fig: whether to return the result of drawing
    :return: DataFrame of the evaluation indicator
    '''
    adata = sc.read_h5ad(eval_path)
    ctrl_key = key_dic['ctrl_key']
    stim_key = key_dic['stim_key']
    pred_key = key_dic['pred_key']
    cell_type_key = key_dic['cell_type_key']
    condition_key = key_dic['condition_key']
    types = list(set(adata.obs[cell_type_key]))
    types.sort()
    degs_list = []
    r2_mean_list = []
    r2_std_list = []
    fig_list = []
    dist_list = []
    for cell_type in types:
        print(cell_type)
        eval_adata = adata[(adata.obs[cell_type_key] == cell_type)]
        r2_mean, r2_std = get_pearson2(eval_adata, key_dic)
        r2_mean_list.append(r2_mean.values.tolist())
        r2_std_list.append(r2_std.values.tolist())
        sc.tl.pca(eval_adata)
        fig = sc.pl.pca(eval_adata, color=condition_key, frameon=False, show=False,
                        title="PCA of " + cell_type + " by Condition", return_fig=True)
        fig_list.append(fig)
        sc.tl.rank_genes_groups(eval_adata, groupby=condition_key, reference=ctrl_key, method="wilcoxon")
        degs_pred = eval_adata.uns["rank_genes_groups"]["names"][pred_key]
        degs_ctrl = eval_adata.uns["rank_genes_groups"]["names"][stim_key]
        common_degs = list(set(degs_ctrl[0:100]) & set(degs_pred[0:100]))
        common_nums = len(common_degs)
        degs_list.append(common_nums)
        print(common_nums)
        r2mean = draw_reg_plot(eval_adata=eval_adata,
                               cell_type=cell_type,
                               reg_type='mean',
                               axis_keys={"x": pred_key, "y": stim_key},
                               condition_key=condition_key,
                               gene_draw=degs_ctrl[:10],
                               top_gene_list=degs_ctrl[:100],
                               save_path=None,
                               title=None,
                               show=False,
                               return_fig=True,
                               fontsize=12
                               )
        fig_list.append(r2mean[2])

        r2var = draw_reg_plot(eval_adata=eval_adata,
                              cell_type=cell_type,
                              reg_type='var',
                              axis_keys={"x": pred_key, "y": stim_key},
                              condition_key=condition_key,
                              gene_draw=degs_ctrl[:10],
                              top_gene_list=degs_ctrl[:100],
                              save_path=None,
                              title=None,
                              show=False,
                              return_fig=True,
                              fontsize=12
                              )
        fig_list.append(r2var[2])
        for i in range(3):
            fig = plt.figure()
            ax = fig.add_axes([0, 0, 1, 1])
            gene = degs_ctrl[i]

            sc.pl.violin(eval_adata, keys=gene, groupby=condition_key, ax=ax, show=False, )
            ax.set_title('Violin Plot of {} in {}'.format(gene, cell_type))
            fig_list.append(fig)
        dist = get_wasserstein_distance(eval_adata, case_key=stim_key, pred_key=pred_key, top_genes=degs_ctrl[:100])
        dist_list.append(dist)
    data = np.array([degs_list,
                     [i[0] for i in r2_mean_list],
                     [i[2] for i in r2_mean_list],
                     [i[1] for i in r2_mean_list],
                     [i[3] for i in r2_mean_list],
                     [i[0] for i in dist_list],
                     [i[1] for i in dist_list],
                     [i[0] for i in r2_std_list],
                     [i[2] for i in r2_std_list],
                     [i[1] for i in r2_std_list],
                     [i[3] for i in r2_std_list]])
    df = pd.DataFrame(data.T,
                      columns=['common DEGs of top 100 genes',
                               'mean R2 of all genes expression mean', 'mean R2 of top 100 genes expression mean',
                               'mean R2 of all genes expression var', 'mean R2 of top 100 expression var',
                               'Wasserstein distance of all genes', 'Wasserstein distance of top 100 genes',
                               'std R2 of all genes expression mean', 'std R2 of top 100 genes expression mean',
                               'std R2 of all genes expression var', 'std R2 of top 100 expression var'],
                      index=types)
    if save_path:
        df.to_csv(save_path + '/log/{}.csv'.format(data_name))
    if return_fig and save_path:
        with PdfPages(save_path + '/images/{}.pdf'.format(data_name)) as pdf:
            for i in range(len(fig_list)):
                pdf.savefig(figure=fig_list[i], dpi=200, bbox_inches='tight')
                plt.close()
    return df


def get_pearson2(eval_adata, key_dic, n_degs=100, sample_ratio=0.8, times=100):
    """
    Calculating the regression coefficient between the predicted perturbation response and the true response
    :param eval_adata: adata of gene expression containing ctrl,stim, pred
    :param key_dic: dictionary of keywords for the dataset, such as
               {'condition_key': 'condition',
               'cell_type_key': 'cell_type',
               'ctrl_key': 'control',
               'stim_key': 'stimulated',
               'pred_key': 'pred',
               }
    :param n_degs: number of differentially expressed genes used
    :param sample_ratio: sampling rate
    :param times: Sampling times
    :return: DataFrame of the square of pearson's coefficient
    """
    stim_key = key_dic['stim_key']
    pred_key = key_dic['pred_key']
    ctrl_key = key_dic['ctrl_key']
    condition_key = key_dic['condition_key']
    sc.tl.rank_genes_groups(eval_adata, groupby=condition_key, reference=ctrl_key, method="wilcoxon")
    degs = eval_adata.uns["rank_genes_groups"]["names"][stim_key][:n_degs]
    df_stim = eval_adata[(eval_adata.obs[condition_key] == stim_key)].to_df()
    df_pred = eval_adata[(eval_adata.obs[condition_key] == pred_key)].to_df()
    data = np.zeros((times, 4))
    for i in range(times):
        stim = df_stim.sample(frac=sample_ratio, random_state=i)
        pred = df_pred.sample(frac=sample_ratio, random_state=i)
        stim_mean = stim.mean().values.reshape(1, -1)
        pred_mean = pred.mean().values.reshape(1, -1)
        stim_var = stim.var().values.reshape(1, -1)
        pred_var = pred.var().values.reshape(1, -1)
        r2_mean = (np.corrcoef(stim_mean, pred_mean)[0, 1]) ** 2
        r2_var = (np.corrcoef(stim_var, pred_var)[0, 1]) ** 2
        stim_degs_mean = stim.loc[:, degs].mean().values.reshape(1, -1)
        pred_degs_mean = pred.loc[:, degs].mean().values.reshape(1, -1)
        stim_degs_var = stim.loc[:, degs].var().values.reshape(1, -1)
        pred_degs_var = pred.loc[:, degs].var().values.reshape(1, -1)
        r2_degs_mean = (np.corrcoef(stim_degs_mean, pred_degs_mean)[0, 1]) ** 2
        r2_degs_var = (np.corrcoef(stim_degs_var, pred_degs_var)[0, 1]) ** 2
        data[i, :] = [r2_mean, r2_var, r2_degs_mean, r2_degs_var]
    df = pd.DataFrame(data, columns=['r2_all_mean', 'r2_all_var', 'r2_degs_mean', 'r2_degs_var'])
    r2_mean = df.mean(axis=0)
    r2_std = df.std(axis=0)
    return r2_mean, r2_std


def draw_reg_plot(eval_adata,
                  cell_type,
                  reg_type='mean',
                  axis_keys={"x": "pred", "y": "stimulated"},
                  condition_key='condition',
                  gene_draw=None,
                  top_gene_list=None,
                  save_path=None,
                  title=None,
                  show=True,
                  return_fig=False,
                  fontsize=14
                  ):
    '''
    Plot the regression line between the predicted perturbation response and the true response
    :param eval_adata: adata of gene expression containing ctrl,stim, pred
    :param cell_type: cell type of the data
    :param reg_type: mean or var
    :param axis_keys: {"x": "pred", "y": "stimulated"}, maybe you need change the values of keys
    :param condition_key: 'condition' by default
    :param gene_draw: a list of genes which you want to color red
    :param top_gene_list: a list of top DEGs whose regression coefficients are to be calculated
    :param save_path: the path where you save the regression figure
    :param title: the title of the the regression figure
    :param show: whether to show the figure
    :param return_fig: whether to return the figure
    :param fontsize: the fontsize of the figure
    :return:
    '''
    df_case = eval_adata[(eval_adata.obs[condition_key] == axis_keys["y"])].to_df()
    df_pred = eval_adata[(eval_adata.obs[condition_key] == axis_keys["x"])].to_df()
    if reg_type == 'mean':
        mean_case = df_case.mean().values.reshape(-1, 1)
        mean_pred = df_pred.mean().values.reshape(-1, 1)
    elif reg_type == 'var':
        mean_case = df_case.var().values.reshape(-1, 1)
        mean_pred = df_pred.var().values.reshape(-1, 1)
    data = np.hstack((mean_case, mean_pred))
    data_df = pd.DataFrame(data, columns=['case', 'predict'], index=df_case.columns)

    fig, ax = plt.subplots()
    sns.set(color_codes=True)
    sns.regplot(x='case', y='predict', data=data_df, ax=ax)
    if gene_draw is not None:
        texts = []
        x = mean_case
        y = mean_pred
        for i in gene_draw:
            j = eval_adata.var_names.tolist().index(i)
            x_bar = x[j]
            y_bar = y[j]
            texts.append(plt.text(x_bar, y_bar, i, fontsize=11, color="black"))
            ax.plot(x_bar, y_bar, "o", color="red", markersize=5)
        adjust_text(
            texts,
            x=x,
            y=y,
            ax=ax,
            arrowprops=dict(arrowstyle="->", color="grey", lw=0.5),
            force_points=(0.0, 0.0),
        )
    if top_gene_list is not None:
        data_deg = data_df.loc[top_gene_list, :]
        r_top = round(data_deg['case'].corr(data_deg['predict'], method='pearson'), 3)
        xt = 0.1 * np.max(data_df['case'])
        yt = 0.85 * np.max(data_df['predict'])
        ax.text(xt, yt, s='$R^2_{top 100 genes}$=' + str(round(r_top * r_top, 3)), fontsize=fontsize, color='black')
    r = round(data_df['case'].corr(data_df['predict'], method='pearson'), 3)
    xt = 0.1 * np.max(data_df['case'])
    yt = 0.75 * np.max(data_df['predict'])
    ax.text(xt, yt, s='$R^2_{all genes}$=' + str(round(r * r, 3)), fontsize=fontsize, color='black')
    if title:
        plt.title(title)
    else:
        plt.title('The Linear Regression of True and Predict Expression ' + reg_type + ' of ' + cell_type)
    if save_path is not None:
        plt.savefig(save_path)
        plt.close()
    if show:
        plt.show()
        plt.close()
    if return_fig:
        return [round(r * r, 3), round(r_top * r_top, 3), fig]
    else:
        return [round(r * r, 3), round(r_top * r_top, 3)]


def get_wasserstein_distance(eval_adata, case_key='stimulated', pred_key='pred', 
                             top_genes=None, cal_type='sum'):
    """
    '''
    This function is used to calculate the Wasserstein distance between the predicted response and the real response
    :param eval_adata: adata of gene expression containing ctrl,stim, pred
    :param case_key: key of perturbed condition
    :param pred_key: key of predictive condition
    :param top_genes: a list of top DEGs whose Wasserstein distance are to be calculated
    :param cal_type: 'sum' or 'mean'
    :return: the Wasserstein distance between the predicted response and the real response
    '''
    """
    dist_list = []
    dist_list_top = []
    pred = eval_adata[(eval_adata.obs["condition"] == pred_key)].to_df()
    case = eval_adata[(eval_adata.obs["condition"] == case_key)].to_df()

    for i in range(pred.shape[1]):
        gene_pred = pred.iloc[:, i].values
        gene_case = case.iloc[:, i].values
        dist = wasserstein_distance(gene_pred, gene_case)
        dist_list.append(dist)

    if top_genes is None:
        res = None
        if cal_type == 'mean':
            res = np.mean(dist_list)
        elif cal_type == 'sum':
            res = np.sum(dist_list)
        return res
    else:
        for gene in top_genes:
            gene_pred = pred.loc[:, gene].values
            gene_case = case.loc[:, gene].values
            dist = wasserstein_distance(gene_pred, gene_case)
            dist_list_top.append(dist)
        res = None
        if cal_type == 'mean':
            res = [np.mean(dist_list), np.mean(dist_list_top)]
        elif cal_type == 'sum':
            res = [np.sum(dist_list), np.sum(dist_list_top)]
        return res


def cluster_evaluation(adata_obs, label_key, cluster_key):
    """
    Evaluation of clustering effect
    :param adata_obs: the observation of the adata
    :param label_key: the key of true label in adata_obs
    :param cluster_key: the key of cluster label in adata_obs
    :return: clustering effect evaluation indicators including AMI, ARI, NMI, HOM, completeness, vms, FMS
    """
    print("clustered by:", cluster_key)
    ARI = sklearn.metrics.adjusted_rand_score(adata_obs[cluster_key], adata_obs[label_key])
    AMI = sklearn.metrics.adjusted_mutual_info_score(adata_obs[label_key], adata_obs[cluster_key])
    NMI = sklearn.metrics.normalized_mutual_info_score(adata_obs[cluster_key], adata_obs[label_key])
    HOM = sklearn.metrics.homogeneity_score(adata_obs[label_key], adata_obs[cluster_key])
    completeness = sklearn.metrics.completeness_score(adata_obs[label_key], adata_obs[cluster_key])
    vms = sklearn.metrics.v_measure_score(adata_obs[label_key], adata_obs[cluster_key])
    FMS = sklearn.metrics.fowlkes_mallows_score(adata_obs[label_key], adata_obs[cluster_key])

    print('AMI:%.3f\tARI:%.3f\tNMI:%.3f\tHOM:%.3f\tcompleteness:%.3f\tV-measure:%.3f\tFMS:%.3f\t' % (
        AMI, ARI, NMI, HOM, completeness, vms, FMS))
    return AMI, ARI, NMI, HOM, completeness, vms, FMS



import scanpy as sc
import numpy as np


def adata_process(adata, min_genes=200, min_cells=10, max_value=10, n_top_genes=6000):
    '''
    Process the raw count matrix of the input
    :param adata: Primitive counting matrix
    :param min_genes: Each cell has at least m genes expressed, otherwise they are filtered
    :param min_cells: Each gene is expressed in at least n cells, otherwise it is filtered
    :param max_value: The maximum value to which the gene expression value is normalized
    :param n_top_genes: The maximum number of differential genes retained
    :return:
    '''
    sc.pp.filter_cells(adata, min_genes=min_genes)
    sc.pp.filter_genes(adata, min_cells=min_cells)
    sc.pp.normalize_per_cell(adata)
    sc.pp.log1p(adata)
    sc.pp.scale(adata, max_value=max_value)
    sc.pp.highly_variable_genes(adata, n_top_genes=n_top_genes)
    adata = adata[:, adata.var.highly_variable]
    df = adata.to_df().clip(lower=0)
    adata.X = df
    return adata


# 按照一定的比例将基因表达矩阵的部分非零值置为0
def add_mask(adata, fraction):
    '''
    Setting the non-zero value of the gene expression matrix to 0 in a certain proportion
    :param adata: adata of input
    :param fraction: the proportion that is set to 0
    :return: adata processed
    '''
    df = adata.to_df()
    idx = df.index
    for i in range(len(df)):
        non_zero_indices = df.iloc[i].index[df.iloc[i].values != 0]
        num_to_zero = int(len(non_zero_indices) * fraction)
        indices_to_zero = np.random.choice(non_zero_indices, size=num_to_zero, replace=False)
        df.loc[idx[i], indices_to_zero] = 0
    adata.X = df
    return adata


from setuptools import setup, find_packages
import pathlib

here = pathlib.Path(__file__).parent.resolve()

setup(
    name="scpram",
    version="0.0.2",
    description="scPRAM accurately predicts single-cell gene expression perturbation response based on attention mechanism",
    long_description="With the rapid advancement of single-cell sequencing technology, we are gradually enabled to delve into the cellular re-sponses to various external perturbations at the gene expression level. However, obtaining perturbed samples in certain scenarios may be considerably challenging, and the substantial costs associated with sequencing also curtail the feasibil-ity of large-scale experimentation. A repertoire of methodologies has been employed for forecasting perturbative re-sponses in single-cell gene expression. However, existing methods only focus on predictions at the average level and do not capture the heterogeneity at the single-cell resolution effectively. Here we present scPRAM, a method for predicting Perturbation Responses in single-cell gene expression based on At-tention Mechanisms. Leveraging variational autoencoders and optimal transport, scPRAM aligns cell states before and after perturbation, followed by accurate prediction of gene expression responses to perturbations for unseen cell types through attention mechanisms. Experiments on multiple real perturbation datasets involving drug treatments and bacteri-al infections demonstrate that scPRAM attains heightened accuracy in perturbation prediction across cell types, species, and individuals, surpassing existing methodologies. Furthermore, scPRAM demonstrates outstanding capability in identi-fying differentially expressed genes under perturbation, capturing heterogeneity in perturbation responses across spe-cies, and maintaining stability in the presence of data noise and sample size variations.",
    license="MIT Licence",
    url="https://github.com/jiang-q19/scPRAM",
    author="Qun Jiang",
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    keywords="single cell, perturbation, attention, optimal transport",
    packages=find_packages(),
    python_requires=">=3.8",
    install_requires=[
        'scanpy==1.9.3',
        'adjusttext==0.7',
        'pot==0.9.1',
        'jupyter==1.0.0',
        'torch==1.13.1',
        'torchaudio==0.13.1',
        'torchvision==0.14.1', 
    ]
)